@@ -1,7 +1,7 @@
 import inspect
 import re
 from functools import wraps
-from typing import Union, Any
+from typing import Union, Any, List
 from uuid import uuid4
 
 import pandas
@@ -11,8 +11,8 @@ from .six import string_types, integer_types
 
 from .fields import (FIELD_DATAFRAME, FIELD_TEXT, FIELD_NUMERIC, FIELD_NO_INPUT,
                      FIELD_SELECT, FIELD_SELECT_MULTIPLE)
-from .utils import fn_name_to_pretty_label, float_to_decimal, vectorized_is_valid, vectorized_date_component, \
-    vectorized_is_complete_date, vectorized_len, vectorized_get_dict_key
+from .utils import fn_name_to_pretty_label, float_to_decimal, vectorized_is_valid, vectorized_compare_dates, \
+    vectorized_is_complete_date, vectorized_len, vectorized_get_dict_key, vectorized_is_in, vectorized_case_insensitive_is_in
 from decimal import Decimal, Inexact, Context
 import operator
 import numpy as np
@@ -263,7 +263,7 @@ class DataframeType(BaseType):
     name = "dataframe"
 
     def __init__(self, data):
-        self.value = self._assert_valid_value_and_cast(data["value"])
+        self.value: pd.DataFrame = self._assert_valid_value_and_cast(data["value"])
         self.column_prefix_map = data.get("column_prefix_map", {})
         self.relationship_data = data.get("relationship_data", {})
         self.value_level_metadata = data.get("value_level_metadata", [])
@@ -301,6 +301,9 @@ class DataframeType(BaseType):
         else:
             return self.value.get(comparator, comparator)
 
+    def is_column_of_iterables(self, column):
+        return isinstance(column, pandas.core.series.Series) and (isinstance(column.iloc[0], list) or  isinstance(column.iloc[0], set))
+
     @type_operator(FIELD_DATAFRAME)
     def exists(self, other_value):
         target_column = self.replace_prefix(other_value.get("target"))
@@ -311,13 +314,12 @@ class DataframeType(BaseType):
         return not self.exists(other_value)
     
     @type_operator(FIELD_DATAFRAME)
-    def equal_to(self, other_value):
+    def equal_to(self, other_value) -> pd.Series:
         target = self.replace_prefix(other_value.get("target"))
         value_is_literal = other_value.get("value_is_literal", False)
         comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
         comparison_data = self.get_comparator_data(comparator, value_is_literal)
-        results = np.where(self.value[target] == comparison_data, True, False)
-        return pd.Series(results)
+        return self.value[target].eq(comparison_data) & ~self.value[target].isin(["", None])
 
     @type_operator(FIELD_DATAFRAME)
     def equal_to_case_insensitive(self, other_value):
@@ -326,8 +328,7 @@ class DataframeType(BaseType):
         comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
         comparison_data = self.get_comparator_data(comparator, value_is_literal)
         comparison_data = self.convert_string_data_to_lower(comparison_data)
-        results = np.where(self.value[target].str.lower() == comparison_data, True, False)
-        return pd.Series(results)
+        return (self.value[target].str.lower() == comparison_data) & ~self.value[target].isin(["", None])
 
     @type_operator(FIELD_DATAFRAME)
     def not_equal_to_case_insensitive(self, other_value):
@@ -372,14 +373,16 @@ class DataframeType(BaseType):
         comparison_data = self.get_comparator_data(comparator, value_is_literal)
         results = np.where(self.value[target] > comparison_data, True, False)
         return pd.Series(results)
-    
+
     @type_operator(FIELD_DATAFRAME)
     def contains(self, other_value):
         target = self.replace_prefix(other_value.get("target"))
         value_is_literal = other_value.get("value_is_literal", False)
         comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
         comparison_data = self.get_comparator_data(comparator, value_is_literal)
-        if isinstance(comparator, pandas.core.series.Series):
+        if self.is_column_of_iterables(self.value[target]):
+            results = vectorized_is_in(comparison_data, self.value[target])
+        elif isinstance(comparator, pandas.core.series.Series):
             results = np.where(comparison_data.isin(self.value[target]), True, False)
         else:
             results = np.where(self.value[target] == comparison_data, True, False)
@@ -396,7 +399,9 @@ class DataframeType(BaseType):
         comparator = self.replace_prefix(other_value.get("comparator")) if not value_is_literal else other_value.get("comparator")
         comparison_data = self.get_comparator_data(comparator, value_is_literal)
         comparison_data = self.convert_string_data_to_lower(comparison_data)
-        if isinstance(comparator, pandas.core.series.Series):
+        if self.is_column_of_iterables(self.value[target]):
+            results = vectorized_case_insensitive_is_in(comparison_data, self.value[target])
+        elif isinstance(comparator, pandas.core.series.Series):
             results = np.where(comparison_data.isin(self.value[target].str.lower()), True, False)
         else:
             results = np.where(self.value[target].str.lower() == comparison_data, True, False)
@@ -415,7 +420,10 @@ class DataframeType(BaseType):
             # column name provided
             comparator = self.replace_prefix(comparator)
         comparison_data = self.get_comparator_data(comparator, value_is_literal)
-        results = self.value[target].isin(comparison_data)
+        if self.is_column_of_iterables(comparison_data):
+            results = vectorized_is_in(self.value[target], comparison_data)
+        else:
+            results = self.value[target].isin(comparison_data)
         return pd.Series(results)
     
     @type_operator(FIELD_DATAFRAME)
@@ -433,11 +441,15 @@ class DataframeType(BaseType):
             # column name provided
             comparator = self.replace_prefix(comparator)
         comparison_data = self.get_comparator_data(comparator, value_is_literal)
-        if isinstance(comparison_data, pd.core.series.Series):
-            comparison_data = comparison_data.str.lower()
-        results = self.value[target].str.lower().isin(comparison_data)
+        if self.is_column_of_iterables(comparison_data):
+            results = vectorized_case_insensitive_is_in(self.value[target].str.lower(), comparison_data)
+            return pd.Series(results)
+        elif isinstance(comparison_data, pd.core.series.Series):
+            results = self.value[target].str.lower().isin(comparison_data.str.lower())
+        else:
+            results = self.value[target].str.lower().isin(comparison_data)
         return pd.Series(results.values)
-    
+
     @type_operator(FIELD_DATAFRAME)
     def is_not_contained_by_case_insensitive(self, other_value):
         return ~self.is_contained_by_case_insensitive(other_value)
@@ -584,7 +596,7 @@ class DataframeType(BaseType):
         target = self.replace_prefix(other_value.get("target"))
         comparator = self.replace_prefix(other_value.get("comparator"))
         component = other_value.get("date_component")
-        results = np.where(operator(vectorized_date_component(component, self.value[target]), vectorized_date_component(component, self.value.get(comparator, comparator))), True, False)
+        results = np.where(vectorized_compare_dates(component, self.value[target], self.value.get(comparator, comparator), operator), True, False)
         return pd.Series(results)
     
     @type_operator(FIELD_DATAFRAME)
@@ -636,22 +648,48 @@ class DataframeType(BaseType):
         return pd.Series(results)
 
     @type_operator(FIELD_DATAFRAME)
-    def is_unique_relationship(self, other_value):
+    def is_not_unique_relationship(self, other_value) -> pd.Series:
+        """
+        Validates one-to-one relationship between two columns (target and comparator) against a dataset.
+        One-to-one means that a pair of columns can be duplicated but its integrity must not be violated:
+        one value of target always corresponds to one value of comparator. Examples:
+
+        Valid dataset:
+        STUDYID  STUDYDESC
+        1        A
+        2        B
+        3        C
+        1        A
+        2        B
+
+        Invalid dataset:
+        STUDYID  STUDYDESC
+        1        A
+        2        A
+        3        C
+        """
         target = self.replace_prefix(other_value.get("target"))
         comparator = other_value.get("comparator")
         if isinstance(comparator, list):
             comparator = self.replace_all_prefixes(comparator)
         else:
             comparator = self.replace_prefix(comparator)
-        grouped_dict = self.value.groupby([comparator])[target].apply(set).to_dict()
-        results = np.where(
-            vectorized_len(vectorized_get_dict_key(grouped_dict, self.value[comparator])) <= 1, True, False
-        )
-        return pd.Series(results)
+        df_without_duplicates: pd.DataFrame = self.value.drop_duplicates()  # remove repeating rows
+        # we need to check if ANY of the columns (target or comparator) is duplicated
+        duplicated_comparator: pd.Series = df_without_duplicates[comparator].duplicated(keep=False)
+        duplicated_target: pd.Series = df_without_duplicates[target].duplicated(keep=False)
+        result = pd.Series([False] * len(self.value))
+        if duplicated_comparator.any():
+            duplicated_comparator_values = set(df_without_duplicates[duplicated_comparator][comparator])
+            result += self.value[comparator].isin(duplicated_comparator_values)
+        if duplicated_target.any():
+            duplicated_target_values = set(df_without_duplicates[duplicated_target][target])
+            result += self.value[target].isin(duplicated_target_values)
+        return result
 
     @type_operator(FIELD_DATAFRAME)
-    def is_not_unique_relationship(self, other_value):
-        return ~self.is_unique_relationship(other_value)
+    def is_unique_relationship(self, other_value) -> pd.Series:
+        return ~self.is_not_unique_relationship(other_value)
 
     @type_operator(FIELD_DATAFRAME)
     def is_not_unique_set(self, other_value):
